// Code generated by stringer -i a.out.go -o anames.go -p loong64; DO NOT EDIT.

package loong64

import "cmd/internal/obj"

var Anames = []string{
	obj.A_ARCHSPECIFIC: "ABSD",
	"ABSF",
	"ADD",
	"ADDD",
	"ADDF",
	"ADDU",
	"ADDW",
	"AND",
	"BEQ",
	"BGEZ",
	"BLEZ",
	"BGTZ",
	"BLTZ",
	"BFPF",
	"BFPT",
	"BNE",
	"BREAK",
	"CMPEQD",
	"CMPEQF",
	"CMPGED",
	"CMPGEF",
	"CMPGTD",
	"CMPGTF",
	"LU12IW",
	"LU32ID",
	"LU52ID",
	"PCALAU12I",
	"PCADDU12I",
	"JIRL",
	"BGE",
	"BLT",
	"BLTU",
	"BGEU",
	"DIV",
	"DIVD",
	"DIVF",
	"DIVU",
	"DIVW",
	"LL",
	"LLV",
	"LUI",
	"MOVB",
	"MOVBU",
	"MOVD",
	"MOVDF",
	"MOVDW",
	"MOVF",
	"MOVFD",
	"MOVFW",
	"MOVH",
	"MOVHU",
	"MOVW",
	"MOVWD",
	"MOVWF",
	"MUL",
	"MULD",
	"MULF",
	"MULU",
	"MULH",
	"MULHU",
	"MULW",
	"NEGD",
	"NEGF",
	"NEGW",
	"NEGV",
	"NOOP",
	"NOR",
	"OR",
	"REM",
	"REMU",
	"RFE",
	"SC",
	"SCV",
	"SGT",
	"SGTU",
	"SLL",
	"SQRTD",
	"SQRTF",
	"SRA",
	"SRL",
	"ROTR",
	"SUB",
	"SUBD",
	"SUBF",
	"SUBU",
	"SUBW",
	"DBAR",
	"SYSCALL",
	"TEQ",
	"TNE",
	"WORD",
	"XOR",
	"MASKEQZ",
	"MASKNEZ",
	"MOVV",
	"SLLV",
	"SRAV",
	"SRLV",
	"ROTRV",
	"DIVV",
	"DIVVU",
	"REMV",
	"REMVU",
	"MULV",
	"MULVU",
	"MULHV",
	"MULHVU",
	"ADDV",
	"ADDVU",
	"SUBV",
	"SUBVU",
	"TRUNCFV",
	"TRUNCDV",
	"TRUNCFW",
	"TRUNCDW",
	"MOVWU",
	"MOVFV",
	"MOVDV",
	"MOVVF",
	"MOVVD",
	"ORN",
	"ANDN",
	"AMSWAPB",
	"AMSWAPH",
	"AMSWAPW",
	"AMSWAPV",
	"AMCASB",
	"AMCASH",
	"AMCASW",
	"AMCASV",
	"AMADDW",
	"AMADDV",
	"AMANDW",
	"AMANDV",
	"AMORW",
	"AMORV",
	"AMXORW",
	"AMXORV",
	"AMMAXW",
	"AMMAXV",
	"AMMINW",
	"AMMINV",
	"AMMAXWU",
	"AMMAXVU",
	"AMMINWU",
	"AMMINVU",
	"AMSWAPDBB",
	"AMSWAPDBH",
	"AMSWAPDBW",
	"AMSWAPDBV",
	"AMCASDBB",
	"AMCASDBH",
	"AMCASDBW",
	"AMCASDBV",
	"AMADDDBW",
	"AMADDDBV",
	"AMANDDBW",
	"AMANDDBV",
	"AMORDBW",
	"AMORDBV",
	"AMXORDBW",
	"AMXORDBV",
	"AMMAXDBW",
	"AMMAXDBV",
	"AMMINDBW",
	"AMMINDBV",
	"AMMAXDBWU",
	"AMMAXDBVU",
	"AMMINDBWU",
	"AMMINDBVU",
	"EXTWB",
	"EXTWH",
	"CLOW",
	"CLOV",
	"CLZW",
	"CLZV",
	"CTOW",
	"CTOV",
	"CTZW",
	"CTZV",
	"REVBV",
	"REVB2W",
	"REVB4H",
	"REVB2H",
	"REVH2W",
	"REVHV",
	"BITREV4B",
	"BITREV8B",
	"BITREVW",
	"BITREVV",
	"BSTRINSW",
	"BSTRINSV",
	"BSTRPICKW",
	"BSTRPICKV",
	"PRELD",
	"PRELDX",
	"CRCWBW",
	"CRCWHW",
	"CRCWWW",
	"CRCWVW",
	"CRCCWBW",
	"CRCCWHW",
	"CRCCWWW",
	"CRCCWVW",
	"RDTIMELW",
	"RDTIMEHW",
	"RDTIMED",
	"CPUCFG",
	"FMADDF",
	"FMADDD",
	"FMSUBF",
	"FMSUBD",
	"FNMADDF",
	"FNMADDD",
	"FNMSUBF",
	"FNMSUBD",
	"FMINF",
	"FMIND",
	"FMAXF",
	"FMAXD",
	"FMAXAF",
	"FMAXAD",
	"FMINAF",
	"FMINAD",
	"FCOPYSGF",
	"FCOPYSGD",
	"FSCALEBF",
	"FSCALEBD",
	"FLOGBF",
	"FLOGBD",
	"FCLASSF",
	"FCLASSD",
	"FFINTFW",
	"FFINTFV",
	"FFINTDW",
	"FFINTDV",
	"FTINTWF",
	"FTINTWD",
	"FTINTVF",
	"FTINTVD",
	"FTINTRPWF",
	"FTINTRPWD",
	"FTINTRPVF",
	"FTINTRPVD",
	"FTINTRMWF",
	"FTINTRMWD",
	"FTINTRMVF",
	"FTINTRMVD",
	"FTINTRZWF",
	"FTINTRZWD",
	"FTINTRZVF",
	"FTINTRZVD",
	"FTINTRNEWF",
	"FTINTRNEWD",
	"FTINTRNEVF",
	"FTINTRNEVD",
	"VMOVQ",
	"XVMOVQ",
	"VADDB",
	"VADDH",
	"VADDW",
	"VADDV",
	"VADDQ",
	"XVADDB",
	"XVADDH",
	"XVADDW",
	"XVADDV",
	"XVADDQ",
	"VSUBB",
	"VSUBH",
	"VSUBW",
	"VSUBV",
	"VSUBQ",
	"XVSUBB",
	"XVSUBH",
	"XVSUBW",
	"XVSUBV",
	"XVSUBQ",
	"VADDBU",
	"VADDHU",
	"VADDWU",
	"VADDVU",
	"VSUBBU",
	"VSUBHU",
	"VSUBWU",
	"VSUBVU",
	"XVADDBU",
	"XVADDHU",
	"XVADDWU",
	"XVADDVU",
	"XVSUBBU",
	"XVSUBHU",
	"XVSUBWU",
	"XVSUBVU",
	"VANDB",
	"VORB",
	"VXORB",
	"VNORB",
	"XVANDB",
	"XVORB",
	"XVXORB",
	"XVNORB",
	"VANDV",
	"VORV",
	"VXORV",
	"VNORV",
	"VANDNV",
	"VORNV",
	"XVANDV",
	"XVORV",
	"XVXORV",
	"XVNORV",
	"XVANDNV",
	"XVORNV",
	"VPCNTB",
	"VPCNTH",
	"VPCNTW",
	"VPCNTV",
	"XVPCNTB",
	"XVPCNTH",
	"XVPCNTW",
	"XVPCNTV",
	"VSEQB",
	"XVSEQB",
	"VSEQH",
	"XVSEQH",
	"VSEQW",
	"XVSEQW",
	"VSEQV",
	"XVSEQV",
	"VDIVB",
	"VDIVH",
	"VDIVW",
	"VDIVV",
	"VDIVBU",
	"VDIVHU",
	"VDIVWU",
	"VDIVVU",
	"VMODB",
	"VMODH",
	"VMODW",
	"VMODV",
	"VMODBU",
	"VMODHU",
	"VMODWU",
	"VMODVU",
	"XVDIVB",
	"XVDIVH",
	"XVDIVW",
	"XVDIVV",
	"XVDIVBU",
	"XVDIVHU",
	"XVDIVWU",
	"XVDIVVU",
	"XVMODB",
	"XVMODH",
	"XVMODW",
	"XVMODV",
	"XVMODBU",
	"XVMODHU",
	"XVMODWU",
	"XVMODVU",
	"VSLLB",
	"VSLLH",
	"VSLLW",
	"VSLLV",
	"VSRLB",
	"VSRLH",
	"VSRLW",
	"VSRLV",
	"VSRAB",
	"VSRAH",
	"VSRAW",
	"VSRAV",
	"VROTRB",
	"VROTRH",
	"VROTRW",
	"VROTRV",
	"XVSLLB",
	"XVSLLH",
	"XVSLLW",
	"XVSLLV",
	"XVSRLB",
	"XVSRLH",
	"XVSRLW",
	"XVSRLV",
	"XVSRAB",
	"XVSRAH",
	"XVSRAW",
	"XVSRAV",
	"XVROTRB",
	"XVROTRH",
	"XVROTRW",
	"XVROTRV",
	"VILVLB",
	"VILVLH",
	"VILVLW",
	"VILVLV",
	"VILVHB",
	"VILVHH",
	"VILVHW",
	"VILVHV",
	"XVILVLB",
	"XVILVLH",
	"XVILVLW",
	"XVILVLV",
	"XVILVHB",
	"XVILVHH",
	"XVILVHW",
	"XVILVHV",
	"VMULB",
	"VMULH",
	"VMULW",
	"VMULV",
	"VMUHB",
	"VMUHH",
	"VMUHW",
	"VMUHV",
	"VMUHBU",
	"VMUHHU",
	"VMUHWU",
	"VMUHVU",
	"XVMULB",
	"XVMULH",
	"XVMULW",
	"XVMULV",
	"XVMUHB",
	"XVMUHH",
	"XVMUHW",
	"XVMUHV",
	"XVMUHBU",
	"XVMUHHU",
	"XVMUHWU",
	"XVMUHVU",
	"VFSQRTF",
	"VFSQRTD",
	"VFRECIPF",
	"VFRECIPD",
	"VFRSQRTF",
	"VFRSQRTD",
	"XVFSQRTF",
	"XVFSQRTD",
	"XVFRECIPF",
	"XVFRECIPD",
	"XVFRSQRTF",
	"XVFRSQRTD",
	"VADDF",
	"VADDD",
	"VSUBF",
	"VSUBD",
	"VMULF",
	"VMULD",
	"VDIVF",
	"VDIVD",
	"XVADDF",
	"XVADDD",
	"XVSUBF",
	"XVSUBD",
	"XVMULF",
	"XVMULD",
	"XVDIVF",
	"XVDIVD",
	"VFCLASSF",
	"VFCLASSD",
	"XVFCLASSF",
	"XVFCLASSD",
	"VFRINTRNEF",
	"VFRINTRNED",
	"VFRINTRZF",
	"VFRINTRZD",
	"VFRINTRPF",
	"VFRINTRPD",
	"VFRINTRMF",
	"VFRINTRMD",
	"VFRINTF",
	"VFRINTD",
	"XVFRINTRNEF",
	"XVFRINTRNED",
	"XVFRINTRZF",
	"XVFRINTRZD",
	"XVFRINTRPF",
	"XVFRINTRPD",
	"XVFRINTRMF",
	"XVFRINTRMD",
	"XVFRINTF",
	"XVFRINTD",
	"VNEGB",
	"VNEGH",
	"VNEGW",
	"VNEGV",
	"XVNEGB",
	"XVNEGH",
	"XVNEGW",
	"XVNEGV",
	"VMULWEVHB",
	"VMULWEVWH",
	"VMULWEVVW",
	"VMULWEVQV",
	"VMULWODHB",
	"VMULWODWH",
	"VMULWODVW",
	"VMULWODQV",
	"VMULWEVHBU",
	"VMULWEVWHU",
	"VMULWEVVWU",
	"VMULWEVQVU",
	"VMULWODHBU",
	"VMULWODWHU",
	"VMULWODVWU",
	"VMULWODQVU",
	"XVMULWEVHB",
	"XVMULWEVWH",
	"XVMULWEVVW",
	"XVMULWEVQV",
	"XVMULWODHB",
	"XVMULWODWH",
	"XVMULWODVW",
	"XVMULWODQV",
	"XVMULWEVHBU",
	"XVMULWEVWHU",
	"XVMULWEVVWU",
	"XVMULWEVQVU",
	"XVMULWODHBU",
	"XVMULWODWHU",
	"XVMULWODVWU",
	"XVMULWODQVU",
	"VMULWEVHBUB",
	"VMULWEVWHUH",
	"VMULWEVVWUW",
	"VMULWEVQVUV",
	"VMULWODHBUB",
	"VMULWODWHUH",
	"VMULWODVWUW",
	"VMULWODQVUV",
	"XVMULWEVHBUB",
	"XVMULWEVWHUH",
	"XVMULWEVVWUW",
	"XVMULWEVQVUV",
	"XVMULWODHBUB",
	"XVMULWODWHUH",
	"XVMULWODVWUW",
	"XVMULWODQVUV",
	"VSHUF4IB",
	"VSHUF4IH",
	"VSHUF4IW",
	"VSHUF4IV",
	"XVSHUF4IB",
	"XVSHUF4IH",
	"XVSHUF4IW",
	"XVSHUF4IV",
	"VSETEQV",
	"VSETNEV",
	"VSETANYEQB",
	"VSETANYEQH",
	"VSETANYEQW",
	"VSETANYEQV",
	"VSETALLNEB",
	"VSETALLNEH",
	"VSETALLNEW",
	"VSETALLNEV",
	"XVSETEQV",
	"XVSETNEV",
	"XVSETANYEQB",
	"XVSETANYEQH",
	"XVSETANYEQW",
	"XVSETANYEQV",
	"XVSETALLNEB",
	"XVSETALLNEH",
	"XVSETALLNEW",
	"XVSETALLNEV",
	"LAST",
}
