// Copyright 2025 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

//go:build !unix || wasm

package tls

func pauseProcess() {
	panic("-wait-for-debugger not supported on this OS")
}
