// Code generated by "stringer -type=SymKind"; DO NOT EDIT.

package sym

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[Sxxx-0]
	_ = x[STEXT-1]
	_ = x[STEXTFIPSSTART-2]
	_ = x[STEXTFIPS-3]
	_ = x[STEXTFIPSEND-4]
	_ = x[STEXTEND-5]
	_ = x[SELFRXSECT-6]
	_ = x[SMACHOPLT-7]
	_ = x[STYPE-8]
	_ = x[SSTRING-9]
	_ = x[SGOSTRING-10]
	_ = x[SGOFUNC-11]
	_ = x[SGCBITS-12]
	_ = x[SRODATA-13]
	_ = x[SRODATAFIPSSTART-14]
	_ = x[SRODATAFIPS-15]
	_ = x[SRODATAFIPSEND-16]
	_ = x[SRODATAEND-17]
	_ = x[SFUNCTAB-18]
	_ = x[SELFROSECT-19]
	_ = x[STYPERELRO-20]
	_ = x[SSTRINGRELRO-21]
	_ = x[SGOSTRINGRELRO-22]
	_ = x[SGOFUNCRELRO-23]
	_ = x[SGCBITSRELRO-24]
	_ = x[SRODATARELRO-25]
	_ = x[SFUNCTABRELRO-26]
	_ = x[SELFRELROSECT-27]
	_ = x[SMACHORELROSECT-28]
	_ = x[STYPELINK-29]
	_ = x[SITABLINK-30]
	_ = x[SSYMTAB-31]
	_ = x[SPCLNTAB-32]
	_ = x[SFirstWritable-33]
	_ = x[SBUILDINFO-34]
	_ = x[SFIPSINFO-35]
	_ = x[SELFSECT-36]
	_ = x[SMACHO-37]
	_ = x[SMACHOGOT-38]
	_ = x[SWINDOWS-39]
	_ = x[SELFGOT-40]
	_ = x[SNOPTRDATA-41]
	_ = x[SNOPTRDATAFIPSSTART-42]
	_ = x[SNOPTRDATAFIPS-43]
	_ = x[SNOPTRDATAFIPSEND-44]
	_ = x[SNOPTRDATAEND-45]
	_ = x[SINITARR-46]
	_ = x[SDATA-47]
	_ = x[SDATAFIPSSTART-48]
	_ = x[SDATAFIPS-49]
	_ = x[SDATAFIPSEND-50]
	_ = x[SDATAEND-51]
	_ = x[SXCOFFTOC-52]
	_ = x[SBSS-53]
	_ = x[SNOPTRBSS-54]
	_ = x[SLIBFUZZER_8BIT_COUNTER-55]
	_ = x[SCOVERAGE_COUNTER-56]
	_ = x[SCOVERAGE_AUXVAR-57]
	_ = x[STLSBSS-58]
	_ = x[SXREF-59]
	_ = x[SMACHOSYMSTR-60]
	_ = x[SMACHOSYMTAB-61]
	_ = x[SMACHOINDIRECTPLT-62]
	_ = x[SMACHOINDIRECTGOT-63]
	_ = x[SFILEPATH-64]
	_ = x[SDYNIMPORT-65]
	_ = x[SHOSTOBJ-66]
	_ = x[SUNDEFEXT-67]
	_ = x[SDWARFSECT-68]
	_ = x[SDWARFCUINFO-69]
	_ = x[SDWARFCONST-70]
	_ = x[SDWARFFCN-71]
	_ = x[SDWARFABSFCN-72]
	_ = x[SDWARFTYPE-73]
	_ = x[SDWARFVAR-74]
	_ = x[SDWARFRANGE-75]
	_ = x[SDWARFLOC-76]
	_ = x[SDWARFLINES-77]
	_ = x[SDWARFADDR-78]
	_ = x[SSEHUNWINDINFO-79]
	_ = x[SSEHSECT-80]
}

const _SymKind_name = "SxxxSTEXTSTEXTFIPSSTARTSTEXTFIPSSTEXTFIPSENDSTEXTENDSELFRXSECTSMACHOPLTSTYPESSTRINGSGOSTRINGSGOFUNCSGCBITSSRODATASRODATAFIPSSTARTSRODATAFIPSSRODATAFIPSENDSRODATAENDSFUNCTABSELFROSECTSTYPERELROSSTRINGRELROSGOSTRINGRELROSGOFUNCRELROSGCBITSRELROSRODATARELROSFUNCTABRELROSELFRELROSECTSMACHORELROSECTSTYPELINKSITABLINKSSYMTABSPCLNTABSFirstWritableSBUILDINFOSFIPSINFOSELFSECTSMACHOSMACHOGOTSWINDOWSSELFGOTSNOPTRDATASNOPTRDATAFIPSSTARTSNOPTRDATAFIPSSNOPTRDATAFIPSENDSNOPTRDATAENDSINITARRSDATASDATAFIPSSTARTSDATAFIPSSDATAFIPSENDSDATAENDSXCOFFTOCSBSSSNOPTRBSSSLIBFUZZER_8BIT_COUNTERSCOVERAGE_COUNTERSCOVERAGE_AUXVARSTLSBSSSXREFSMACHOSYMSTRSMACHOSYMTABSMACHOINDIRECTPLTSMACHOINDIRECTGOTSFILEPATHSDYNIMPORTSHOSTOBJSUNDEFEXTSDWARFSECTSDWARFCUINFOSDWARFCONSTSDWARFFCNSDWARFABSFCNSDWARFTYPESDWARFVARSDWARFRANGESDWARFLOCSDWARFLINESSDWARFADDRSSEHUNWINDINFOSSEHSECT"

var _SymKind_index = [...]uint16{0, 4, 9, 23, 32, 44, 52, 62, 71, 76, 83, 92, 99, 106, 113, 129, 140, 154, 164, 172, 182, 192, 204, 218, 230, 242, 254, 267, 280, 295, 304, 313, 320, 328, 342, 352, 361, 369, 375, 384, 392, 399, 409, 428, 442, 459, 472, 480, 485, 499, 508, 520, 528, 537, 541, 550, 573, 590, 606, 613, 618, 630, 642, 659, 676, 685, 695, 703, 712, 722, 734, 745, 754, 766, 776, 785, 796, 805, 816, 826, 840, 848}

func (i SymKind) String() string {
	if i >= SymKind(len(_SymKind_index)-1) {
		return "SymKind(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _SymKind_name[_SymKind_index[i]:_SymKind_index[i+1]]
}
