// Copyright 2025 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package cgroup

var ContainsCPU = containsCPU

var ParseV1Number = parseV1Number
var ParseV2Limit = parseV2Limit

var ParseCPURelativePath = parseCPURelativePath
var ParseCPUMount = parseCPUMount

var UnescapePath = unescapePath
