// Copyright 2024 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

TEXT	·asm(SB),0,$0-8
	CALL	·callback(SB)
	RET
