// Copyright 2024 The Go Authors. All rights reserved.
// Use of this source code is governed by a BSD-style
// license that can be found in the LICENSE file.

package issue69912

// Define an arbitrary type name, which will be used to demonstrate
// the race of issue #69912.
type T int
